--[[

=== INTRODUCTION ===

    The DevConsole is a developer's utility for logging statements to a visible console and/or to a tab within the current project.  It has many advantages over the built-in Codea console, the primary being that there is more real-estate available so it is easier to view more complex logs.  Other advantages will become apparent when you browse through the available options and features.

    The draw() cycle can use a lot of CPU with too many consoles or too large a buffer in a console.  You should generally not create consoles if you don't need them, or at least hide or minimize them and use the "show on print" or "maximize on print" options to have them pop up only when one of your DevConsole:print() statements creates something worth seeing.  Hidden consoles take up virtually nothing.  Minimized consoles use some fractional CPU time to draw the minimized button, but not much.  The use of a reasonable buffer limit is highly encouraged.

    If you have trouble when using too many consoles... well, don't.  Try using just one, and limit the overall buffer size to something within reason.


=== QUICK START ===

 --- Include DevDevConsole ---

    Include the DevDevConsole project in your own project's dependencies.

    You can quickly see and test its behavior by adding a call to DevConsole.testAll() to the very end of your Main draw() function and DevConsole.touchedAny(touch) at the very beginning of your Main touched(touch) function, and then run your project.  

    This will create several random consoles, and generate random log statements to them.  The testAll() function has parameters that can be used to see and test various different features or feature combinations and behaviors, but that is not relevant for the Quick Start.  Just try it out in the simplest way.

    [Do not use DevConsole:testAll() or DevConsole:touchedAll(); the proper form is DevConsole.testAll() and DevConsole.touchedAny(touch)].  

    PLEASE: Make sure to read the "Polite Suggestions" section of this document before changing settings or moving too far beyond this Quick Start.

    After trying things out this way, remove your call to DevConsole.testAll(), and fix your call to DevConsole.touchedAny(touch) as outlined below.

 --- Create Your Own DevConsole --- 

    Create a console and assign it to a variable so that you can reference it.  You may create more than one console.  You can also use the DevConsole.getConsole() method to find/recover a console Example:

        debugDevConsole = DevConsole()
        performanceDevConsole = DevConsole(0.5, 0.5, 0.5, 1, "Performance", DevConsole.THEME_BATMAN)

    The main numeric parameters are the x,y screen location of the lower left corner of the console, and then the width and height.  Any value between 0 and 1 will be treated as a fraction of the screen width or height, and will respond appropriately when the device orientation changes.

    The first string given will be a specific title for the console.

 --- Call DevConsole.drawAll() and DevConsole.touchedAny() ---
                
    [ If you added any code for the Quick Start, remove it now. ]

    Add a call to DevConsole.drawAll() to your own Main draw() funtion, preferably at the very end, so consoles are drawn on top of everything you draw.

    Add a call to DevConsole.touchedAny(touch) to you Main touched(touch) function, almost certainly at the very beginning, since consoles will generally "float on top of" any touchable elements of your own.  DevConsole.touchedAny() will return true if it handled a touch, and if it does, you usually don't want to let your own code handle that same touch; you should exit the function immediately when DevConsole.touchedAny() returns true.  

    For example:

        function draw()
            -- Your draw code here
            DevConsole.drawAll()
        end

        function touched(touch)
            if DevConsole.touchedAny(touch) then return end
            -- Your touch code here
        end

    NOTE: These are not "instance" methods, they are "static" functions... They must be called using DevConsole-dot , not yourDevConsole-colon.

 --- Call yourDevConsole:print() ---

    Write to the console in the same way that you would use a Lua print() statement, but referencing the particular console instance:

        debugDevConsole:print("How did the code get here?")
        performanceDevConsole:print("Elapsed Time:",ElapsedTime)
        sillyConsole:print(color(255,0,255)," is almost like octarine")

    You can print lines in another color to highlight important lines:

        debugDevConsole:printWithColor(color(255,0,0),"RED ALERT!")
        debugDevConsole:printWithColor(true, "A color of 'true' uses the highlight color for your console, which defaults to yellow")

 --- Using The DevConsole Interface ---

    Each console has buttons and drag points with which you can adjust the console.

    -- Buttons ---

    There are buttons at the top, to the left and right of the console title.  The buttons are:
        +    Zoom in (make the console text larger)
        -    Zoom out (make the console test smaller)
        ◎    Reset the zoom (font size) to the default
        ⇩    Save the current log contents to a tab (file) within the current project (for reference after terminating your program)
        ✄   Clear the console (remove all accumulated log statements)
        ⎘    Copy the console to the clipboard (so you can paste it into another app, as uncolored text)
        X    Minimize the console; this moves the console to a small button at a corner of your screen, where it can be pressed to expand again

    --- Touches ---

    You can also:

        Pinch to zoom in and out of text (change the font size)
        Use one finger to scroll (up/down, or if necessary, left/right)
        Drag from the left, right or bottom margins, or any corner, to resize the console UNLESS you have set the fixedSize option to true
        Drag the console by its title area to reposition it UNLESS you have set the fixedPosition option to true
        Tap once to bring the console to the front
        Double-tap to make the console the size of the entire screen, or restore it to its previous size
        Triple-tap to send the console behind all other consoles
        Quadruple-tap to display all console settings in the console itself
        Long-press the title (without dragging) to hide the console; THE ONLY WAY to redisplay it once hidden is to call DevConsole:show() method or set showOnPrint to true (and print to it)

    --- Destroying Consoles ---

    If you no longer need a console, you should call the :close() method.  If you don't a copy of the console will remain in memory, referenced by the internal DevConsole.CONSOLES table used to draw consoles, with all of the memory it consumes (it's entire buffer).  It will also continue to be drawn, unless you have hidden it.

=== CUSTOMIZATION === 

 --- Best Practices ---

    DevConsole has many customizable options.  The easiest way to customize a console is to include parameters on construction.  Alternately, you can use the setter methods to set them after creating the console.  DO NOT SET INTERNAL DevConsole fields yourself.  Always use a setter.  Even if it seems to work, it might not, either now or in the future.

    If you find yourself preferring certain options all the time, you could override the provided DevConsole.DEFAULT_... constants.

    IF YOU DO... my advice is to create your own console project ("MyDevConsole"), have that project include the DevConsole project as a dependency, and in your project reset any already defined DevConsole.DEFAULT_... values to whatever you wish.  Then make your own console project the dependency of any projects in which you use it.  The advantage to this approach is that if I distribute a new version with bug fixes or new features, it will be very easy for you to drop it into place and take advantage of the changes.  Your only task would be to make sure that your constant overrides remain compatible with the new version.

    Of course, an equivalent (and probably safter) method is just to create your own function to create consoles with your preferences (rather than override the DevConsole.DEFAULT_ constants), and then use that function to create your consoles.


=== FEATURES AND SETTING OPTIONS ===

 --- Creating DevConsoles with Options ---

    Options can be set when a console is created by passing the parameters either in order (which is difficult beyond the first, most common few), or by passing a table with named (and, optionally, sometimes unnamed) parameters.

    The DevConsole uses a special method to evaluate parameters.  

    If you simply pass parameters, they will be interpreted in the order they are expected, by the type of parameter that applies.  This means that all of the following work:

        wordsDevConsole = DevConsole(200, 200, 400, 400, "Words", color(255), color(0, 150)-- x, y, w, h, title, text color, background color
        bugsDevConsole = DevConsole("Bugs", color(0,255,0), color(0,50,0,150), 0.25, 0.25, 0.5, 1) -- title, text color, background color, x, y, w, h
        goofyDevConsole = DevConsole(0.75, color(255,0,255), "I Am Goofy", 0.25, 0.25, color(50,0,50,150), 0.5, 0.5) -- x, text color, title, y, w, background color, h... this is silly, but it works   

    If you pass parameters as a table, parameters will be taken from the table.  The first priority will be to use the parameter by name.  If not available by name, parameters will be selected in order by type:

        namesDevConsole = DevConsole{title="Names",0.5, 0, 0.5, 0.5, "Courier", textColor=color(255), bgColor=color(0,0,150,180)}

    Lastly, you can also list tables in tables (to any depth) as parameters.  The primitive components (numbers, strings, booleans) will be added to the parameter list as if they had been encountered in the outer table.  This is what allows the DevConsole.THEME_ constants to work, and can by used by developers to create option sets of their own.

    Note that named parameters that come later in the list will override the previous values.

 --- Themes (Color Sets) ---

    Many themes have been established as constants.  To use them, simply pass the entire color set as an unnamed parameter, like:

        DevConsole{DevConsole.THEME_DRACULA,0,0,WIDTH/2,HEIGHT} -- DevConsole on the left half of the display with the "Dracula" theme

    You can look at the code to see the theme available, and you can always override specific colors if you prefer, either by using a setter after initialization, or by adding an override after the included theme in the parameter list.

        DevConsole{DevConsole.THEME_SUPERMAN, highlightColor=color(0,255,0)} -- Superman theme using kryptonite to highlight text

 --- Setting Options On The Fly ---

    Settings can be made after a console is created, and can even be made much later, after the console as already been in use, by using the provided setter methods (DevConsole:setWhatever).  Every setter is named in camelcase, using the defined paremeter name, such as:

        myDevConsole:setTitle("I Have Changed")
        myDevConsole:setX(0.1)
        myDevConsole:setTextColor(color(255,255,0,200))

 --- Print Constants ---

    Print constants are special values that impact the behavior of a particular print statement.  They are passed as arguments to the print statement (along with the values to be printed), but are not themselves printed.  They are recognized as "print constants" by the print() method and are interpreted to change the behavior of the print().

    Two methods exist to allow the developer to create one's own print constants.  

    createPrintColor() will create and return a constant that represents a particular color for the console, referencing either an explicit color an index to a color for a console's theme.

    createPrintSeparator() will create and return a constant that represets a separator (a string) to place between each value in the print statement.

    NOTE: It is certainly possible, and even recommended, to set your own constants to these values, to shorten them.  For example:

        local HILITE = DevConsole.PRINT_HIGHLIGHT
        local COMMAS = DevConsole.PRINT_COMMA_SPACE_SEPARATOR
        local TABLES = DevConsole.PRINT_EXPAND_TABLES
        myConsole:print(HILITE, COMMAS, TABLES, a, b, c)

    The available, built-in print constants are:

        DevConsole.PRINT_COMMA_SEPARATOR        Uses a comma to separate print elements
        DevConsole.PRINT_COMMA_SPACE_SEPARATOR  Uses a comma, then a space, to separate print elements
        DevConsole.PRINT_EXPAND_TABLES          Prints the table ID, and then every key=value pair in the table.
                                                NOTE: This option is not recursive.  That is, if a table contains a table, the key-value pairs of that embedded table are NOT expanded.
                                                NOTE: It might be useful to also include the PRINT_NEWLINE_SEPARATOR when this constant is used.
        DevConsole.PRINT_HIGHLIGHT              Uses the highlight color to color that line
        DevConsole.PRINT_NEWLINE_SEPARATOR      Uses a new line to separatoe print elements
        DevConsole.PRINT_NO_SEPARATOR           Uses nothing (an empty string) to separate print elements (just like the Lua print() function)
        DevConsole.PRINT_SPACE_SEPARATOR        Uses a space to separate print elements
        DevConsole.PRINT_TAB_SEPARATOR          Uses a tab to separate print elements
        DevConsole.PRINT_TEXT_COLOR             Uses the text color to print the line 
                                                NOTE: There's really no reason I can think of to ever use this, unless you are programmatically setting colors, and need a value to revert to the text color.

    NOTE: A print line can only be a single color, so if multiple color constants are passed to a print() method, only the last one supplied is used.

    NOTE: A print line can use multiple seprators.  If a separator constant is encountered, all elements after that will use that separator, possibly until a different separator is specified.

 --- Copy and Save ---

    One useful feature is the ability to save the contents of the console to a tab in your project.  Once you terminate your program, the built-in Codea console is gone, along with any errors or output you have sent to it, so you can only view it while the program is running.

    In DevConsole, you can copy the contents of the console to the pasteBoard using the copy button (or by calling the copyToPasteboard() method), allowing you to paste it into any application you like for review or searching, while running or after terminating your program.

    Alternately, DevConsole will allow you to save the contents of the buffer to a tab in your project.

    TO MAKE THE SAVE-TO-TAB BUTTON AVAILABLE ON A CONSOLE, you must specify the tabName property.  You also need that property declared in order to call the :saveToTab() method for a console without specifying the tab name as a parameter (although you can provide the tab name at the time you make the call, even if a default tabName value is not set).

    NOTE: When saving to a new tab, that tab will not become visible until you close and reopen your project (at least, that's the Codea behavior as of 03-09-2022).

    By saving the console to a tab, you can view it after you have terminated your program, conveniently from within Codea, and you can use the Codea Find functionality to search it.

    There are two important options when saving to a tab.  One is to require that the tab be empty.  This is provided so that you cannot accidentally overwrite any of your existing code.  With this option set, if the tab you name is not empty, a new tab will be created with a _# suffix.  For example, if the tab MySave is not empty, and the tab MySave_1 is not empty, the save will be placed into a new MySave_2 tab.

    The other option is to append the result to the tab.  This is the default setting, and guarantees that at worst you will accidentally append the console contents (which are written as Lua comments) to any code.

    IT'S NOT MY FAULT OR MY PROBLEM if you use this improperly and wind up writing over your own code.  It works fine, but pay attention to what you are doing. 

    NOTE: Currently, the display filter is NOT applied when doing a copy or save.  The entire unfiltered contents of the buffer are included.

    [Creating a distinct setCopySaveFilter is on my TODO list!]

 --- Programmatic Control ---

    Before listing the available options and settings, there are certain methods that might be called programmatically, for instance to make the consoles respond to your own controls, or to control the visibility of consoles based on your own logic.  This is a list of such methods, which are documented in greater detail further down in this document:

        DevConsole:disable() -- Disables the console, so print all statements are ignored
        DevConsole:enable() -- Enables the console, so print statements work
        DevConsole:hide() -- Hides the console completely (and saves your CPU from any effort to draw it at all)
        DevConsole:show() -- Shows (but does not maximize) a hidden console
        DevConsole:minimize() -- Minimizes a console (turning it into a button at a corner of your screen)
        DevConsole:maximize() -- Shows (if hidden) and expands a minimized console to show its contents
        DevConsole:clear() -- Clears all content from the console
        DevConsole:toBack() -- Sends the console behind (when drawing) all other consoles
        DevConsole:toFront() -- Brings the console in front of all other consoles
        DevConsole:saveToTab() -- Writes the contents of the console to a named tab in your project
        DevConsole:copyToPasteboard() -- Copies the contents of the console to the clipboard
        DevConsole:getAsString() -- Get the contents of the console as a string
        DevConsole:setPrintFilter() -- Sets a filter function that prevents certain lines from being printed
        DevConsole:setDisplayFilter() -- Sets a filter function that hides certain lines
        DevConsole:setHighlightFilter() -- Sets a filter function that automatically highlights certain lines
        DevConsole:setColorFilter() -- Sets a filter function that colors qualifying lines with a specified color
        DevConsole:clearColorFilters() -- Clears any existing color filters
        DevConsole:printSettings() -- Writes all of the console parameter settings to the console itself
        DevConsole:restoreToOrigin() -- Reset a console to its original position and size (after dragging/resizing)
        DevConsole.restoreAll() -- A "static" function, moves all consoles back to their original positions and sizes
        
        copyToPasteboard() and getAsString take the same parameters:

            fromLine -- A starting line number to extract (ignore lines before that line number)
            endLine -- An ending line number to extract (ignore lines after that line number)
            timestamp -- Set to true or false to override the console's printTime option for this operation
            lineNumbers -- Set to true or false to override the console's lineNumbers option for this operation

        All filter functions return true if the line qualifies (should print/display/highlight/color) or false otherwise.  The setColorFilter takes a color before the filter parameters.  Each takes the one or more filter parameters that will be interpreted as follows:

            -- A string, containing a pattern, that will be applied to the line of text (but not the line number or time stamp)
            -- A number, meaning to print/display from that time or line number on; values of seven digits or more are treated as times
            -- A table consisting of two numbers, specifying the range of lines or times (inclusive) to print
            -- A function that will receive the entire line buffer, and return true if that line qualifies [See below for a discussion of the line buffer]
            -- A table (list) of strings, numbers, number-pairs (tables), or functions, interpreted as above; if any returns true, then the line qualifies

        When more than one filter parameter is specified, if any result qualifies, the line qualifies.  There is no "and/or" logic available.

        Any filter can be cleared by calling the set method without any parameters.

        The line buffer is a table consisting of the following values, any of which can be used by the filter:

            --  str : the line of text to be displayed
            --  col : the color to use to display the line
            --  lin : the line number
            --  tim : the time the line was written (in iOS time, i.e. Unix epoch time)
            --  cnt : the number of consecutive times this text was printed (NOTE: Only one line exists in the buffer for the text, and so filtering the line filters ALL copies of the line)
            --  lst : the last time that this line was printed (relevant for printing consecutive copies of the same line)

        NOTE: In theory you can change the buffer contents as well, and while this is not advised, there should be no ill-effects in most cases.

 --- Feature and Option List ---
        
    Options are listed here first by category.  Beyond this list is an alphabetic list where you can look up the impact of each option.

    --- By Category ---

        --- Display Format ---

            bgColor
            borderColor
            borderWidth
            firstLine
            fontName
            fontSize
            margin
            title
            titleColor
            titleSize
            :zoomIn()
            :zoomOut()
            :zoomReset()

        ---  Display Visibility ---

            hidden
            :hide()
            hideWhenEmpty
            :isBack()
            :isFront()
            :maximize()
            maxOnPrint
            :minimize()
            minimizeAcross
            minimizeBottom
            minimizeLeft
            minimizeRight
            minimizeTop
            minimizeUpDown
            minimized
            minimizedLabel
            :show()
            showOnPrint
            :toBack()
            :toFront()

        --- Printing ---

            bufferLimit
            :clear()
            :createPalette()
            .createPrintColor(colorOrIndex)
            :createPrintColor(colorOrIndex)
            .createPrintSeparator(separator)
            :createPrintSeparator(separator)
            countColor
            :disable()
            :enable()
            enabled
            echo
            highlightColor
            lineGap
            :lock()
            :logPalette()
            numberFormat
            :print(...)
            :printc(...)
            :printcs(...)
            :printIf(condition, ...)
            :printIfc(condition, lineColor, ...)
            :printIfWithColor(condition, lineColor, ...)
            :prints(...)
            :printt(...)
            :printWithColor(lineColor, ...)
            printLineNumbers
            printSeparator
            :printSettings()
            printTime
            printSeparator
            textColor
            timeFormat
            :unlock()
            wrapLogLines

        --- Filtering ---

            :clearColorFilters()
            :setColorFilter(lineColor, ...)
            :setDisplayFilter(...)
            :setHighlightFilter(...)
            :setPrintFilter(...)

        --- Size and Position ---

            fixedPosition
            fixedSize
            h
            :restoreToOrigin()
            x
            y
            w

        --- Saving and Copying

            appendToTab
            :copyToPasteboard(fromLine, toLine, printTime, printLineNumbers)
            saveHighlightChar
            :saveToTab(tabName)
            tabName
            toEmptyTabOnly

        --- General ---

            :close()
            .drawAll()
            :getAsString(fromLine, toLine, printTime, printLineNumbers)
            .getConsole(parm)
            .testAll(consoles, parms, delay)
            .touchedAny(touch)
        
    --- Settings ---
                    
        appendToTab / getAppendToTab / setAppendToTab

            Type: Boolean
            Default: true

            Set to true to cause the Save To Tab button or any saveToTab() call to append the console contents to the tab, rather than replacing its contents.

            It is HIGHLY recommended that you leave this as true, to avoid any chance whatsoever of the function destroying your code by accident.

        bgColor / getBgColor / setBgColor

            Type: Color
            Default: dark, translucent gray

            The color of the background of the console.

        borderColor / getBorderColor / setBorderColor

            Type: Color
            Default: The same color as the title

            The color of the border around the console, the buttons in the console, and the minimized representation of the console.

        borderWidth / getBorderWidth / setBorderWidth

            Type: Number
            Default: 2

            The width (in pixels) of the borders drawn around consoles and buttons.

        bufferLimit / getBufferLimit / setBufferLimit

            Type: Number
            Default: 100

            The number of distinct console lines that will be retained at any time.

            Set this to 0 to have it retain all lines.  This setting is HIGHLY DISCOURAGED!!!

        countColor / getCountColor / setCountColor

            Type: Color
            Default: White -- Color(255)

            The color of the count displayed when consecutive logged lines are repeated.

        echo / getEcho / setEcho 

            Type: Boolean
            Default: false

            Whether (true) or not print statements to the console should also be sent to the Codea console.

        enabled / setEnabled / getEnabled / :enable() / :disable()

            Type: Boolean
            Default: true

            Whether (true) or not print statements should be sent to the console.  When disabled, all print statements will be ignored.

            WARNING: The only feedback to the developer that a console has been disabled is a single line logged to the console itself.  No feedback is given for individual print statements that are ignored.

        firstLine / getFirstLine / setFirstLine

            Type: String
            Default: nil

            A line of text to immediately print to the console upon creation.

        fixedPosition / getFixedPosition / setFixedPosition

            Type: Boolean
            Default: false

            Whether (true) or not the console can be repositioned by dragging.  Note that you can always resize a console by programmatically setting the x and y values of the console.

        fixedSize / getFixedSize / setFixedSize

            Type: Boolean
            Default: false

            Whether (true) or not the console can be resized by dragging.  Note that you can always resize a console by programmatically setting the width or height.

        fontName / getFontName / setFontName

            Type: String
            Default: Inconsolata (a simple, monospace, sans serif font)

            The font to use for all text in the console, including the title, buttons and printed lines.

        fontSize / getFontSize / setFontSize

            Type: Number
            Default: 12

            The font size to use for printed lines in the console when not zoomed in or out.  Note that you can reset the console to this font size by calling :zoomReset().

        h / getH / setH

            Type: Number
            Default: 1 -- means the entire HEIGHT of the display

            The height of the console.  A number between 0 and 1 is treated as a fraction of the current HEIGHT of the device, and will automatically adjust as the orientation of the device changes.  Values greater than 1 are treated as a count in pixels.

        hidden / getHidden / setHidden

            Type: Boolean
            Default: false

            Whether (true) or not the console is visible.  This property can be set with either :setHidden(), :show() or :hide().

        hideWhenEmpty / getHideWhenEmpty / setHideWhenEmpty
                                    
            Type: Boolean
            Default: false

            Whether (true) or not the console will automatically hide (not minimize, hide) when empty (such as after it is cleared either programmatically or using the clear console button).

        highlightColor / getHighlightColor / setHighlightColor

            Type: Color
            Default: Yellow -- color(255, 255, 0)

            The color to use to highlight lines (i.e. when calling printWithColor with the first parameter 'true' rather than a color).

        lineGap / getLineGap / setLineGap

            Type: Number
            Default: 4

            The number of pixels of space to place between each print line.  This helps distinguish the end of a wrapped printed line from the next line.

        margin / getMargin / setMargin

            Type: Number
            Deafult: 20

            The number of pixels of inset space within the console, between any elements (title, buttons, content) and the outside border.

        maxOnPrint / getMaxOnPrint / setMaxOnPrint

            Type: Boolean
            Default: false

            Whether (true) or not to maximize a minimized console when you print to it.

        minimizeAcross / getMinimizeAcross / setMinimizeAcross

            Type: Boolean
            Default: false

            Whether, when minimized, to stack this console horizontally across the screen (true) or vertically (false).

        minimizeBottom / getMinimizeBottom / setMinimizeBottom

            Type: Boolean
            Default: true

            Whether, when minimized, the minimized console should be placed at the bottom (true) or top (false) of the screen.

        minimizeLeft / getMinimizeLeft / setMinimizeLeft

            Type: Boolean
            Default: false

            Whether, when minimized, the minimized console should be placed at the left (true) or right (false) of the screen.

        minimizeRight / getMinimizeRight / setMinimizeRight

            Type: Boolean
            Default: true

            Whether, when minimized, the minimized console should be placed at the right (true) or left (false) of the screen.

        minimizeTop / getMinimizeTop / setMinimizeTop

            Type: Boolean
            Default: false

            Whether, when minimized, the minimized console should be placed at the top (true) or bottom (false) of the screen.

        minimizeUpDown / getMinimizeUpDown / setMinimizeUpDown

            Type: Boolean
            Default: true

            Whether, when minimized, to stack this console vertically (true) or horizontally across the screen (false).

        minimized / getMinimized / setMinimized

            Type: Boolean
            Default: false

            Whether (true) or not the console is currently minimized.

        minimizedLabel / getMinimizedLabel / setMinimizedLabel

            Type: String
            Default: The first letter of the title of the console, or the number of the console if no explicit title was given

            The label (a single character is recommended) that will identify the console in minimized form.

        numberFormat / getNumberFormat / setNumberFormat

            Type: String (a string.format() formatting value)
            Default: %4.3f   IMPORTANT: If you don't want your values truncated this way, set numberFormat to nil
        
            The format to use whenever printing number values to the console.

            IMPORTANT: This defaluts to three decimmals, under the assumption that most people do not care about precision beyond that for normal applications.  If this is incorrect in your circumstances, you have two options, either set this value to something else (nil or an empty string to have no formatting applied), or call string.format() yourself in your print statement, so the DevConsole sees the exact string you want to print, rather than a number.

        palette / getPalette / setPalette

            Type: Table
            Default: A table of colors created from the text and highlight colors

            A table of a selection of colors to use for printing lines distinguished from the text and highlight colors.  Entry 1 is always the text color, entry 2 is the highlight color.

        printLineNumbers / getPrintLineNumbers / setPrintLineNumbers

            Type: Boolean
            Default: true

            Whether to print line numbers (true) on the console, or not. 

        printSeparator / getPrintSeparator / setPrintSeparator

            Type: String
            Default: "" (nothing, an empty string)

            The value to print between values in a print statement.

            The default behaves like the traditional Lua print() statement, i.e. all values are strung together.

            If set, this value will be effect for all subsequent print() statements of any kind, unless an explicit PRINT_..._SEPARATOR constant is used.  For more information, refer to the section on Print Constants.

        printTime / getPrintTime / setPrintTime

            Type: Boolean
            Default: true

            Whether to print the time each line was logged (true) on the console, or not.

        saveHighlightChar / getSaveHighlightChar / setSaveHighlightChar

            Type: String
            Default: •

            The character to include in the pasteboard or saved-tab text versions of the log.  Note that several DevConsole.ICON_ constants exist with alternate characters that you might choose to make use.

        showOnPrint / getShowOnPrint / setShowOnPrint

            Type: Boolean
            Default: true

            Whether to show a hidden (not minimized, hidden) console when a new line is printed to it (true), or to leave it hidden.

        tabName / getTabName / setTabName

            Type: String
            Default: None (an empty string)

            The name to use for the tab in your project where the console contents can be saved.  NOTE: YOU MUST SPECIFY THIS VALUE FOR THE FUNCTIONALITY TO WORK, and for the save-to-tab button to be included in the console.

        textColor / getTextColor / setTextColor

            Type: Color
            Default: Slightly transparent white -- color(255,255,255,150)

            The color of normal, printed text in the console.

        timeFormat / getTimeFormat / setTimeFormat

            Type: String (a valid strformat value)
            Default: "%M:%S"

            The format to use to print log times.  The default was chosen as the best to use in day-to-day debugging, where your program will run for at most a few minutes.  That default exists as the constant DevConsole.ALT_TIMESTAMP_MINSECS

            Another valuable option, perticularly where performance is an issue, is to use "%f", which displays the time in microseconds.  This constant exists as DevConsole.ALT_TIMESTAMP_MICROSECS.

            For cases where you may be saving and referring to logs for longer periods, you may want the full date, using "%y %m %d %T".  This constant exists as DevConsole.ALT_TIMESTAMP_YMDT

        title / getTitle / setTitle

            Type: String
            Default: Special

            The title of the console.  If none is specified, the console will be named "Console #", where # is the number of the console in order of creation.

            Note that you can give a console "no title" by calling console:setTitle("") after the console is first created.

        titleColor / getTitleColor / setTitleColor

            Type: Color
            Default: The same as the text color

            The color to be used to draw the title of the console.

        titleSize / getTitleSize / setTitleSize

            Type: Number
            Default: 20

            The font size of the title of the console.

        toEmptyTabOnly / getToEmptyTabOnly / setToEmptyTabOnly

            Type: Boolean
            Deafult: false

            Whether to only save the console to an empty tab (true), creating a new tab when necessary, or to use an existing, non-empty tab for the save.

        w / getW / setW

            Type: Number
            Default: 1 -- means the entire WIDTH of the display

            The width of the console.  A number between 0 and 1 is treated as a fraction of the current WIDTH of the device, and will automatically adjust as the orientation of the device changes.  Values greater than 1 are treated as a count in pixels.

        wrapLogLines / getWrapLogLines / setWrapLogLines

            Type: Boolean
            Default: true

            Whether lines in the console should wrap (true) or extend (truncated) beyond the edge of the display.

        x / getX / setX

            Type: Number
            Default: 0

            The x coordinate of the console.  A number between 0 and 1 is treated as a fraction of the current WIDTH of the device, and will automatically adjust as the orientation of the device changes.  Values greater than 1 are treated as a count in pixels.

        y / getY / setY

            Type: Number
            Default: 0

            The y coordinate of the console.  A number between 0 and 1 is treated as a fraction of the current HEIGHT of the device, and will automatically adjust as the orientation of the device changes.  Values greater than 1 are treated as a count in pixels.

    --- Methods ---

        :clear()

            This method clears the buffer for the console.  Note that if hideWhenEmpty is set to true, the console will hide as a result.

        :clearColorFilters()

            Clear all set color filters.

        :close()
            
            Destroy internal references to the console, so it can be garbage collected.
    
        :copyToPasteboard(fromLine, toLine, printTime, printLineNumbers)

            Parameters: See the :getAsString() method

            This method copies the contents of the buffer to the pasteboard.
    
            Highlighted lines are denoted using the string value of saveHightlightChar.

        :createPalette(textColor, highlightColor, bgColor)

            Parameters:
                textColor:          (Color, optional) The text color on which to base the palette.  Does not set .textColor!  Defaults to the console text color.
                highlightColor:     (Color, optional) The highlight color on which to base the palette.  Does not set .highlightColor!  Defaults to the console highlight color.
                bgColor:            (Color, optoinal) The background color which cannot be added to the palette.  Does not set .bgColor!  Defaults to the console background color.

            Creates a new palette based on the given colors, or the current colors assigned to the console.

        .createPrintColor(colorOrIndex)
        :createPrintColor(colorOrIndex)

            Parameters:
                colorOrIndex:       (Color or number) The color to which this print constant refers.

            Creates a "print constant" that can be passed as an argument to any print statement to change the color of that line.  The constant itself is NOT printed.

            When given a number, the constant will represent that color index in whichever console the constant is used (NOT the color from a console that created the index; it represents the index, not a specific color).

            If more than one color print constant is specified, the LAST one supplied will take effect for the entire line (you can't use this to color different parts of the line)

            Note that this method can be called in either the DevConsole.createPrintColor() or anyConsole:createPrintColor() form.

            See the secton on Print Constants for more information.

        .createPrintSeparator(printSeparator)
        :createPrintSeparator(printSeparator)

            Parameters:
                printSeparator:       (String) The separator to use between elements in a print command.

            Creates a "print constant" that can be passed as an argument to any print command to change the separator from that point on in that statement.  Note that more than one value can be passed in one print() command, resulting in different separators at different points.

            Note that this method can be called in either the DevConsole.createPrintSeparator() or anyConsole:createPrintSeparator() form.

            See the secton on Print Constants for more information.

        :disable()

            Disable the console, so that print statements are ignored.  This is useful for incurring the performance hit of logging only when your program is in an execution state where you care.

            NOTE: You can use the Codea console input to manually enable and disable your console.  For example, type:

                myConsole:enable()

        :enable()

            Enable the console, so that print statements work.

        :getAsString(fromLine, toLine, printTime, printLineNumbers)
            
            Parameters:    
                fromLine:           (Number, optional) The lowest line number to copy
                toLine:             (Number, optional) The highest line number to copy
                printTime:          (Boolean, optional) When specified, this overrides the printTime setting for the console for this call.
                printlineNumbers:   (Boolean, optional) When specified, this overrides the printLineNumbers setting for the console for this call.

            This method returns the unfiltered contents of the console as a string (less any pending lines for a locked console).  This method is used by the :copyToPasteboard() and :saveToTab() methods to construct their content.

            Contents are written bracketed by comments, and any "end multi-line string" indicators will be changed to "]\]" to prevent syntax errors in your project.

            Highlighted lines are denoted by the addition of the saveHighlightChar (default is •) before the print time.

            For example:
 
--		 22 03 08 06:15:18	0003	All Angry, Loud Bugs Hate Moldy Rocks 
--	•	 22 03 08 06:15:21	0004	Those Curious, Lazy Gnats Hate Malodorous Puddles But Maybe All Small, Grumpy Gnats Eat Useless Cheese 
--		 22 03 08 06:15:25	0005	The Tiny, Green Eels Hate Insidious Feathers

        .getConsole(parm)

            Parameters:
                parm:               (Number or String) The index of the console in the DevConsole.CONSOLES list or a pattern used to match the title.  See the description for more detailed information on this parameter.

            Finds a console by index or title.

            NOTE: Console index numbers change!  Consoles are maintained in the CONSOLES list in the order they are drawn, from back to front.  Changing the order of the consoles will change their indexes.

            NOTE: A parm value of zero is treated as the "top" of the console list, that is, the console that is at the front.  Negative values count back from the top, that is, -1 is the second console from the front, -2 the third, etc.

            Use :getConsole(1) to get the console that is furthest back.

            Use :getConsole(0) to get the console at the front.

            NOTE: This method is used by both :copyToPasteboard() and :saveToTab().

            NOTE: This method does not (currently) apply any filters.

            NOTE: This method excludes lines printed after the console was locked.

        :getPaletteSize()

            This method returns the size of the palette, i.e. the number of colors available in it (including the text and highlight colors).

        :hide()

            Hides the console.  Hidden consoles can only be redisplayed by calling :show(), or by printing to them when the showOnPrint option is set to true.
            
        :isBack()

            Returns true if the console is the furthest back of all consoles.

        :isFront()

            Returns true if the console is the frontmost of all consoles.
            
        :lock()

            Locks the console, so that print statements are accumulated, but cease to be displayed, copied or saved.  Such lines will appear after the console is unlocked.

        :logPalette()

            Writes a line to the console for each color available in the palette.

        .luaPrint(...)

            This is a saved copy of the Lua/Codea print() function, provided as a convenience for developers that override that function definition.

        :maximize()

            Maximize the console, i.e., restore it to display its contents).

        :minimize()

            Minimize the console, i.e., draw it as a small, square button (press to maximize) at a corner of the screen.

        :print(...)

            Print data to the console.  This method works exactly like the Lua/Codea print() function.

        :printc(lineColor, ...)

            Alias for :printWithColor(lineColor, ...); see below.

        :printcs(...)

            Print data to the console, using the comma-space separator (a convenience method).

        :printIf(condition, ...)

            Parameters:
                condition:      (Boolean) Whether to print this line or not
                
            The line is printed only if condition is true.  While similar to using a printFilter(), this allows you to easily focus a condition on specific print statements (such as when debugging, and only wishing to see the output under certain specific conditions).

        :printIfc(condition, lineColor, ...)

            Alias for :printIfWithColor(lineColor, ...).

        :printIfWithColor(condition, lineColor, ...)

            Parameters:
                condition:          (Boolean, Optional). The line is printed only if this not false or nil
                lineColor:          See :printWithColor()

            Does a :printWithColor(lineColor, ...); see below, but only if the condition is true.  See :printIf() for more explanation.

        :prints(...)

            Print data to the console, using the space separator (a convenience method).

        :printt(...)

            Print data to the console, using the expand tables option (a convenience method).

        :printtc(...)

            Print data to the console, using the expand tables option and the comma separator (a convenience method).

        :printtcs(...)

            Print data to the console, using the expand tables option and the comma-space separator (a convenience method).

        :printtn(...)

            Print data to the console, using the expand tables option and the newline separator (a convenience method).

        :printSettings()

            This method copies all current console settings to the buffer itself.  useful for figuring out what you have or have not configured properly.

        :printWithColor(lineColor, ...)

            Parameters:
                lineColor        : (Color or Boolean). The color to use for the line.  When the boolean true value is used, the defined highlight color is used.

            Print data to the console, using a specified color.

        .restoreAll()

            Show and maximize all consoles.

        :restoreToOrigin()

            Restore this console to its original, defined position and size.

        :saveToTab(tabName)

            Parameters:
                tabName:        (String, optional) The name of the tab to write to.

            Writes the contents of the buffer, unfiltered, to the named tab in your project.  If the tab does not yet exist, it will be created, but you must close and reopen your project to see any new tabs.

            See the :getAsString() method for more information on the construction of the buffer content for character (non-screen) display.

            The addition of print times and line numbers are controlled by the related console options (printTime, printLineNumbers and timeFormat).

        :setColorFilter(lineColor, ...)

            Parameters:

                lineColor:      (Color, Boolean or nil)  The color to use on any qualifying lines.  If true, then the highlight color is used.  If nil AND no filter parameters are given, then all color filters are cleared.  If a color is given without filter parameters, then the filter for that color is cleared.
                ...:            A sequence of filter parameteers.  See the Filter Parameters section for more information.

            This sets a color filter.  Any matching lines will display in the given color.  This filter has no impact on the :copyToPasteboard() or :saveToTab() output.

        :setDisplayFilter(...)

            Parameters:

                ...:            A sequence of filter parameters.  See the Filter Parameters section for more information.

            This sets a filter, so that only qualifying lines are displayed on the screen.  This filter has no impact on the :copyToPasteboard() or :saveToTab() output.

        :setHighlightFilter(...)

                ...:            A sequence of filter parameters.  See the Filter Parameters section for more information.

            This sets a filter, so that qualifying lines are highlighted on the screen using the set hightlightColor option.  This filter has no impact on the :copyToPasteboard() or :saveToTab() output.

        :setPrintFilter(...)

                ...:            A sequence of filter parameters.  See the Filter Parameters section for more information.

            This sets a filter, so that only qualifying lines are added to the console buffer.  Non-qualifying lines never make it to the console buffer, they are never saved, displayed or printed, asn so are "lost".  This filter has no impact on the :copyToPasteboard() or :saveToTab() output.

        :show()

            Show the console, if it is hidden.

        :toBack()

            Move this console to the very back of the consoles displayed.

        :toFront()

            Make this console the frontmost console, that is, it is displayed on top.

        :unlock()

            Unlock this console.  When unlocked, any pending lines that were added to the console after it was locked will be displayed.

            NOTE: There is no buffer limit on pending lines (at the moment), so leaving a console locked while sending a lot of prints to it can cause you memory problems.

        :zoomIn()

            Increase the font size of the console.

        :zoomOut()

            Decrease the font size of the console.

        :zoomReset()

            Reset the font size of the console to the value originally assigned at the time the console was created.  Note that a call to setFontSize() does not change that original value.  The original font size is ONLY set at construction time.

    --- Filter Parameters ---

            Filter parameters are interpreted according to the following rules:

            One of the constants DevConsole.ANY or DevConsole.ALL determines if any or all supplied conditions must be met.  It should appear exactly once in your parameter lisrt, but if you include it more than once, the last occurance is in effect for the entire filter.  By default, "any" is in effect.

                A future enhancement might allow for a nested-table and-or-and structure, with an additional option for negation (e.g. "not containing this word"), however, as the filter will currently take a function as an argument, and that function can be as complex as you wish, this capability does already exist... provided you are willing to write your own filter function (see below for how).

            A string is interpreted as a Lua pattern, to be matched against the text of each console line.

            A color must match the color assigned to each console line.

            A boolean true will match lines with the highlight color, while false will match lines without the highlight color.

            A number is interpreted as meaning "equal to or greater than this value".  If greater than 9999999, the value is treated as an os.time() value (i.e. Unix epoch time) and the print time of the line is evaluated.  If less then 9999999, the value is treated as a line number.

            A pair of numbers is treated as a range (inclusive) of values, interpeted as print times or line numbers, as with a single number value.

            If a function is provided, that function is passed the buffer line to evaluate, and should return true if the line is to be included for the filter.  A function is, of course, able to access any information your provide to it (through globals, or a closure), but the primary intent is to use the line buffer provided.

            The line buffer is a table with the following contents:

                str = the text of the line
                col = the display text color
                lin = the line number
                cnt = the count of the number of consecutive prints of this text and color
                tim = the time this line was added (OS time, i.e. epoch milliseconds)
                lst = the time a duplicate was last added to this line (again, OS time)

            Note: Filter parameters are handled recursively.  A table with two numbers is treated as described, but any other table provided is investigated, and its contents are also included as filter parameters.  For example,

                :getDisplayFilter{23,{"[Bb]ug","[Dd]og"}} is the same as :getDisplayFilter{23,"[Bb]ug","[Dd]og"}

 --- Testing ---

    I created the DevConsole.testAll() method for my own test purposes, but some developers may find it useful to experiment with various settings to evaluate behavior or performance.

    To use testAll, add it as a call *in place of* DevConsole.drawAll(), at the end your Main draw() method.

    testAll() takes up to three parameters that will be used the first time it is called:

        consoles -- The number of consoles to create; the default is 5.  if not a number, this parameter is ignored (or, when false, triggers the destruction of previous test consoles).
        parms -- A table of any DevConsole setup parameters you want used; if a parameter value is a table, one of the values in that table will be randomly selected for that parameter for a particular console instance; if a parameter value is a function, that function will be called and expected to return the value to be used for a particular console instance.  The default is 
        delay -- An optional delay, in seconds, between random log statements sent to the generated test consoles.

    The test consoles can be destroyed by passing false as the first parameter to testAll().

    Test consoles are only created on the first call to testAll(), or the first call after a testAll(false) call is used to destroy previous test consoles.

 -- I Don't Recommend... ---

    -- Overriding Lua's print() function --

    You could, if you wanted, create your own global print() function (overridding the Lua print function) in order to prevent the mistake of calling print() instead of console:print(), out of habit.  But you may want to use the print function

    -- Handling Touches --

    The prescribed method of letting the DevConsole handle touches first, and only taking touches yourself that the DevConsole does not, is based on the presumption that the consoles will "float" on top of your own display, and need to respond to touches within them.

    One could, of course, imagine a world where the code handles touches "through" a console.  This is not recommended (minimize or hide the console when you need to do so)... but if you really, really want to, you could of course handle all touches yourself, first, and then choose whether or not to pass them to DevConsole.touched(touch).  For example, if all you have is a few buttons on your screen, you could first test to see if the first touch was within your own button (whether covered by a translucent console or not), and only give DevConsole the touch if you didn't use it.  This would let you press your buttons through the console, but otherwise scroll and interact with the console as long as your first touch wasn't in one of your own buttons.

    This isn't recommended, but it's your code.  You simply need to understand the interactions and tradeoffs so that what you do is logical and the resulting behavior is understodd rather than frustrating.


=== POLITE SUGGESTIONS ===

    Don't use too many consoles, in order to avoid performance problems in both drawing and touch handling.

    Keep your buffer small (100 lines?) unless you actually need more, in order to avoid performance problems in both drawing and touch handling.

    Use this project as a dependency, rather than embedding a copy of the DevConsole in your project.  This will make it much easier to accept updates with new features or bug fixes.

    Don't change defined constant values in DevConsole itself.  Override them (if you really must) in your own file (which will be applied after the DevConsole is imported as a dependency).

    Create a project template with DevConsole as a dependency (and with any overrides already defined in Main, if you wish), then use that to create new projects.

    Don't turn off both toEmptyTabOnly and appendToTab when using the saveToTab function or button, unless you want to risk accidentally overwriting some of your own, precious code (either through your own fault, or a bug).

    Archive your projects regularly (not only in case you overwrite a source file, but also just because it's a good practice to follow).  To do so:

        1.  In the Projects page, long-press the target project until the menu comes up
        2.  Choose Export...
        3.  Choose Export Zipped Project
        4.  Choose Open With...
        5.  Choose Save To Files...
        6.  RENAME the file to save with the current date (and if necessary a suffix) to make the file name unique (I use yyyy mm dd format for better sorting by name)
        7.  Choose the destination folder (on your iPad, or, preferably, on iCloud or Dropbox)
        8.  Click Save
        9.  Click Done


=== WARRANTY ===

    None.  I did this for myself, and offer it to anyone else to use because I think some people might want to use it... I'll fix bugs and add features, or not, as I please.  Because I don't get paid to do this.  I do it for fun.  Just like you.  Sorry if this sounds grumpy, but the fact is that while most people are nice and polite, there exists a noisy and innevitable few who are seemingly entitled and not-coincidentally ill-behaved, particularly when they are on the Internet.

    In particular, as of 03/10/2022, I have done very little testing other than unit testing individual features, any one or more of which might have broken due to later development.  So... kindly report any issues (see below), but I ask for your patience.


=== CONTACT ===

    If you have any comments or questions, feel free to contact me by DM through codea.io, at the user blacatena.


=== COPYRIGHT AND USER AGREEMENT ===

    I really don't care.  Do whatever you want, as long as it's not something evil.

    -- Bob Lacatena
--]]